/* (c) Joerg Schulenburg 2000-2007

 this is part of spin-package

 it is for division of matrix multiplication 
 in parallelizable block matrix multiplication
 
    b = H * a    (a,b vectors, H matrix)
    
    b[i] = sum_j H[i,j] * a[j]
    
 blocks can be multiplicated step by step on one system or
 they can be multiplicated on different processors/systems
 
 but you have to pay additional network/harddisk traffic
 
 if b_num is one you have only one block (similar to old version)
 
 - use mpi_n nodes and mpt_n (1..B_NUM) threads per node
 - block counts from 0 to mpt_n (B_NUM) in each node


 */

#ifndef _HBLOCK_H
#define _HBLOCK_H
#ifdef CONFIG_MPI
#include <mpi.h>
#endif

#include "config.h"

#ifndef B_NUM
/* maximum number of threads, threads can be set by option -t<num> */
#define B_NUM 32 /* default max num of compute blocks (threads per node) */
#endif

#ifndef F_NUM
#define F_NUM B_NUM      /* later moved to config.h, can be modified */
#endif

/* #include "defs.h" */

extern int mpi_n, mpi_i;	/* number of MPI nodes + nodenr.         */
extern int pt_n;                /* number of threads per node (1..B_NUM) */
/* pointer for blocked and locally stored vectors (MPI!) */
extern unsigned long node_ofs, node_n1;	/* part of space on this node */
/* pointer for blocked and locally stored vectors (NUMA!) [mpi_n*pt_n] */
extern unsigned long *b_ofs;	/* start index 0..n1-1 of block */
extern unsigned long *b_len;	/* block len */

/* this is for stored files, ToDo: move to h_file */
extern unsigned long b_hnz[B_NUM];	/* NumNonZero-H block entries */
/* this is for partly stored files, ToDo: remove or replace */
extern unsigned long b_snz[B_NUM];	/* stored NonZero-H block entries */
extern unsigned long b_sn1[B_NUM];	/* stored block lines */
extern int b_bits;			/* b_len[0..n-2]=1l<<b_bits */
extern unsigned long b_mask;		/* ~((~0)<<b_bits) */
extern unsigned long len_full_block;   /* number of elements per block */

/* b_idx2blk index-to-block (old name b_getnum)
 * calculate blocknumber (0..(pt_n*mpi_n)-1) from idx (0..n1-1)
 *   thread = block % mpi_n;
 *   node   = block / mpi_i;
 * old version of blocked H and new version of striped (blocked within) H
 * some versions before blocked H and striped single CPU versions
 *      are much faster
 * ToDo: if b_ofs[] is not used fully (blocks<B_NUM) we can have trouble
 */
static inline int b_idx2blk(unsigned long idx){	/* do it fast as possible! */
#if 0
  /* searching the table to find the block, slow, bad scaling */
  int a, m, b; /* left border, middle, right border 2.26 Feb07 */
  for (a=0,b=mpi_n*pt_n,m=(a+b)/2; a!=m; m=(a+b)/2) {
    if (idx >= b_ofs[m]) a=m; else b=m;
    /* log2(blocks)*(1add+1shft+2cmp+2jmp) */
  }
  return m;
#else
  return idx/len_full_block; /* div: speed-70% on PeniumM !? but scales */ 
#endif
}

/* initialization, n1 is whole matrix size */
int b_ini(unsigned long n1,int vvv);
void b_free(); /* opposit of b_ini */

#include "h_file.h"
extern struct thxy b_hbuf[B_NUM]; /* buffer for faster access, [HBLen] */
extern int          b_hbl[B_NUM]; /* num valid data in buffer bufh */  

#ifndef CONFIG_MPI
#define MPI_Barrier(x)     /* dummy function if MPI is not used */
#endif

#endif   /* ifndef _HBLOCK_H */
