/*
 *  this is a collection of diagonalization code used by spinpack
 */

#include "mem.h"
#include "spinstri.h"
#include "spins.h"
#include "block.h"
#include "vector.h"
#ifdef CONFIG_PTHREAD
#include <pthread.h>
#endif

zahl ta[MAX_ITR];
zahl tb[MAX_ITR];		/* a,b Tridiagonalmatrix		*/
zahl tc[MAX_ITR][NEV];		/* eigenvectors of tridiagonalmatrix	*/
    
/* --- --- get smallest eigenvalues by Lanczos (Gagliano) ----------   
  - projections vector |Pi> (start vector |P0>)
  - 2nd Ebenenvektor |Pis>=((H-<H>_i)|Pi>)/(<H^2>_i-<H>_i^2)^(1/2)
    perpenticular to |Pi> (<Pi|Pis>=0)
  - compute 2x2-matrix for |Pi> and |Pis> and their EW
  - |Pi+1>=a|Pi>+b|Pis> is normalized EV to lower EW

  v0=|Pi> v1=H|Pi> v2=HH|Pi>
  h=<H>_i hh=<HH>_i hhh=<HHH>_i
  b=q(<H^2>_i-<H>_i^2)=q(hh-h^2)          mit  q(x)=x^(1/2)
  H-Matrix=[[<Pi|H|Pi>,<Pi|H|Pis>],[<Pis|H|Pi>,<Pis|H|Pis>]]
          =[[h,b],[b,(hhh-2*hh*h+h^3)/b^2]]
          =[[h,b],[b,(hhh-h*hh)/b^2-h]]
          =[[h,b],[b,2*b*f+h]]
  EW=(a11+a22)/2-q((a11+a22)^2/4-a11*a22+a12^2)
    =(hhh-h*hh)/(2b^2)-q((hhh-h*hh)^2/(4b^4)-h*((hhh-h*hh)/b^2-h)+b^2)
    =(hhh-h*hh)/(2b^2)-q((hhh-h*hh)^2/(4b^4)-h*(hhh-h*hh)/b^2+hh)
    =h+b*f-q((b*f+h)^2-2*h*b*f-h^2+b^2) ,    a22=2*b*f+h
    =h+b*f-q(b^2*f^2+b^2)
    =h+b*(f-q(1+f*f))                   ,    alf=f-q(1+f*f)
    =h+b*alf
  EV=[(EW-a22)/a12,1]/q((EW-a22)^2/a12^2+1) * [Pi,Pis]
    =((EW-a22)/b*v0+(1/b*v1-h/b*v0))/q((EW-a22)^2/b^2+1)
    =((EW-a22-h)*v0+v1)/q((EW-a22)^2+b^2)
  f=(a22-a11)/(2*a12),     -p/2=(a11+a22)/2=a22-a12*f=a12*f+a11=b*f+h
   --- last change 7.12.98 JS ueberarbeitet ---
 * --- --- -------------------------------------------------- ---- --- */
/* --- Inp: v0  Out: v0,itr,ew  Wrk: v1 */
int lanz1(tvec *v0,tvec *v1,tvec *ev){
    tvec *v2; cplx a,alf,ak,b,f,r,h,hh,hhh;
    int itr,j,blk; ULong i; double eps=1e-6;	
    /* eps=precission^2 = <(H-<H>)^2> = <HH>-<H><H> >= 0, also complex ??? */
    v2=ev;		/* new 12.09.1998 use memory of ev[n] */
    r=ak=alf=f=0;
    for(itr=0;;itr++){
      clrvec(v1);hamilton(v0,v1);		/* v1=H*v0+v1 */
      clrvec(v2);hamilton(v1,v2);
      for(hhh=hh=h=0,blk=0;blk<pt_n;blk++)for(i=0l;i<b_len[blk];i++){
        r=Conj(M_f2d(v0->v[blk][i]));
        h += r*M_f2d(v1->v[blk][i]);
        hh+= r*M_f2d(v2->v[blk][i]);
        hhh+=Conj(M_f2d(v1->v[blk][i]))*M_f2d(v2->v[blk][i]); }
      b=hh-h*h;		/* negativ if rounding-errors! cmplx>0 ??? */
      b=fabs(b);	/* if cmplx (gegen 0) */
      if(D(b)>0){ /* new projection vector perpenticular to the old one */
        b=sqrt(b); 				/* 2x2 diagonalization */
        f=(hhh-3*h*hh+2*h*h*h)/(2*b*b*b);	/* gegen inf */
        alf=f-sqrt(1.+f*f);			/* gegen -0 bis f*f OVL */ 
        a=1./sqrt(1.+alf*alf);			/* gegen  1 */
        ak=a*alf/b;				/* gegen -0 */
        r=a-ak*h;				/* gegen  1 */
        for(blk=0;blk<pt_n;blk++)for(i=0l;i<b_len[blk];i++)
        v0->v[blk][i]=M_d2f(r*M_f2d(v0->v[blk][i])+ak*M_f2d(v1->v[blk][i]));
        /* h=h+b*alf; */ }
      printf("\n i=%3d h=%16.12f alf=%.4e b=%.1e v=[% .3f,% .3f] t+=%s",
             itr,D(h),D(alf),D(b),D(r),D(h*ak),sclock(~0l));
      if(/* itr>1 && */ fabs(b)<=eps) break; }
    for(blk=0;blk<pt_n;blk++)for(i=0l;i<b_len[blk];i++)
      ev->v[blk][i]=v0->v[blk][i];
    for(j=0;j<NEW;j++)ew[j]=D(h);
    return(itr); }
      
/* ----------- bisection method -------------------------------------
 * calculate the eigenvalues of a tridiagonal matrix T=(tb,ta,tb)
 *  Inp: ta,tb=Tridiagonalmatrix(T),  nn=size of the matrix T
 *       ne=number of eigenvalues ew to calculate, ep=precision
 *  Out: ew  eigenvalues of T via "Nullstellen" of Det(A-x*I)
 *  What happens for degeneracy?
 *    => EW only one-fold, but should be n-fold (as expected by vec12)
 *    ToDo: after EW[ne] is converged, call another exact/slow tridiag
 *          p.e. dsterf() of lapack/icepack ? t(300)=126s acceptable
 *          sparc-Ultra1-Boson: time/s = 1.5e-6 dim ^ 3.2 
 *      OR try (modified) vec12 morefold on same eigenvalue until <tc>=0
 *      check on til8_18b.dat
 * --- */
void bisec(int nn,int ne,zahl ep){
  static zahl b2[MAX_ITR]; /* if not enough stackspace, faster */
  zahl ra,rb,rc,g,epa,range; int i,j,k,neg,ipass;
  /* find a interval containing all the eigenvalues rb..ra */
  range=fabs(ta[0])+fabs(tb[0]);
  for(k=1;k<nn-1;k++) range=max(range,fabs(tb[k-1])+fabs(ta[k])+fabs(tb[k]));
  range=max(range,fabs(tb[nn-2])+fabs(ta[nn-1])); range=-range;
  b2[0]=0; for(i=1;i<nn;i++) b2[i]=Norm(tb[i-1]);
  epa=fabs(range*ep); for(i=0;i<ne;i++) ew[i]=-range;
  rb=range;
  /* find the eigenvalues of a tridiagonal matrix * 
   * lying in a given interval rb...ra            */
  for(k=0;k<ne;k++) { ra=ew[k];       /* ra=upper bound, rb=lower bound */
    /* was j<100, but this is a bit to small ??? */
    for(j=0;j<10000;j++){		/* sturm sequencing */
      if(fabs(ra-rb)<epa)break;	/* ra=rb => break */
      rc=(ra+rb)*0.5; ipass=neg=0; g=1.0;
      for(i=0;i<nn;i++){   /* g=Det_{i}(rc)/Det_{i-1}(rc) */
        if(ipass==0)g=rc-ta[i]-b2[i]/g;else	/* do not divide if g=0 */
        if(ipass==1)ipass=2;
        else      { ipass=0; g=rc-ta[i]; }	/* new start, decoupled */
        if(ipass==0){ if(g<0.0)neg++;		/* sign-change of Det */
          if(fabs(g)<=fabs(b2[i]*epa*ep)) ipass=1; } /* g==0 ? */
      }   
      neg=nn-neg;   /* num of nullstellen rc...ra -> rb...rc */
      if(neg<=k)rb=rc;		/* new lower bound */
      else    { ra=rc; 		/* new upper bound */
                for(i=k;i<min(neg,ne);i++)ew[i]=rc; }
    } }
  return; }

/* --- eigenvectors (EV) of a tridiagonal matrix by inverse iteration */
/*  see Numerical Recipes  (Feb99)
 *   A - matrix
 *   a - start-eigenvalue
 *   x - startvector (random)
 *   step1: solve equation system (A-a*1)*y==x for vector y (ex: LU decomposition)
 *   step2: x := y/|y|          (normalize)
 *   step3: a := a+(x,x)/(x,y)  (not necessary every step)
 *   if Ax=ax then break else goto step1   
 *
 * --- nn i Matrixgroesse, {ta,tb} TriDiagMatrix -> tc EVs
 * --- for degeneracy EW must be n-fold in ew[] and nev>=n!  (does not work!)
 *   ex: ew[]={ -3.5, -3.5, -3.5, -1.5} for 3-fold degeneracy
 *   input: nn - number of sites, nev - number eigenvectors,
 *          ne0 - offset for lowest eigenvectors (0..n1-NEV)
 */
zahl di[MAX_ITR],bl[MAX_ITR],bu[MAX_ITR],
     bv[MAX_ITR],cm[MAX_ITR];
char lex[MAX_ITR];	/* 0 (non pivoting) or 1 (pivoting) */
void vec12(int nn, int nev, int ne0){
    zahl s,prd; int i,j,k,km,l;
    km=0;/* km=0 added by JS */
    for(k=0;k<nev;k++){ /* if ew0==ew1==ew2 : 3 orthogonal vectors */
        for(j=0;j<nn;j++) {di[j]=ew[k+ne0]-ta[j]; bl[j]=bu[j]=-tb[j];}
        /* --- LU decomposition, prepare lex, cm, bu, bv, di */ 
        for(j=0;j<nn-1;j++) {
           if(fabs(di[j])>fabs(bl[j])){		/* --- non pivoting */
             lex[j]=0; if(fabs(di[j])<1e-30) di[j]=1e-30; /* e-40 => e-30 (float) */
             cm[j+1]=bl[j]/di[j]; di[j+1]-=cm[j+1]*bu[j]; bv[j]=0; }  
           else {				/* --- pivoting */
             lex[j]=1; cm[j+1]=di[j]/bl[j]; di[j]=bl[j]; s=bu[j];
             bu[j]=di[j+1]; bv[j]=bu[j+1]; di[j+1]=s-cm[j+1]*bu[j];
             bu[j+1]=-cm[j+1]*bv[j]; }
        }  
        if(fabs(di[nn-1])<1e-30) di[nn-1]=1e-30;
        
        /* --- initial vec tc */ 
        for(j=0;j<nn;j++) tc[j][k]=1./((j+1)*5.); /* random ??? */
        /* --- degeneracy check up -> Entartung km..k */
        /*  degeneracy is given in ew[] implicitely!  */
        if(k==0) km=k; else 
         if(fabs(ew[k+ne0]-ew[km+ne0])>1e-10) km=k;     /* old=1e-13 1998 */ 
         else for(i=km;i<k;i++){       /* orthogonalise ini_vec to tri_ev */
            for(prd=0,j=0;j<nn;j++) prd+=tc[j][i]*tc[j][k];  /* <c_i|c_k> */
            for(      j=0;j<nn;j++) tc[j][k]-=prd*tc[j][i]; }
        /* ToDo: stop and ignore EW, if <tc|tc>==0 ? */
        /* --- inverse iteration of tc (inp: lex, cm, bu, bv, di) */
        for(l=0;l<(k-km+3);l++){	/* what does l??? */
            if((l!=0)||(k!=km)){ /* --- forward substitution */
               for(j=0;j<nn-1;j++)
               if(lex[j]==0) tc[j+1][k]-=cm[j+1]*tc[j][k];
               else {s=tc[j][k];tc[j][k]=tc[j+1][k];
                     tc[j+1][k]=s-cm[j+1]*tc[j][k];}
            }  /* --- backward substitution */
            for(j=nn-1;j>=0;j--){
                s=tc[j][k];
                if(j<(nn-1))s-=bu[j]*tc[j+1][k];
                if(j<(nn-2))s-=bv[j]*tc[j+2][k];
                tc[j][k]=s/di[j];
            }   /* --- normalization */
            for(prd=0,j=0;j<nn;j++) prd+=Norm(tc[j][k]);
            if(prd>1e-30)prd=1./sqrt(prd);else prd=0;
            for(j=0;j<nn;j++) tc[j][k]*=prd;
        }
    }   return;
}

/*
 * the degeneracy from ta,tb is smaller than the degeneracy of H
 *         (and the same as got by fulltridiag)
 * probably reason:
 *   we start with a random vector which defines only a subspace if H has
 *    degenerated eigenvalues
 *   the other degenerated states are natural (mostly or always?) orthogonal
 *    to this space by definition
 *   that should be a problem for all iterating algorithms
 */

/* ew  - eigenvalues of iteration i
 * et  - eigenvalues of iteration i-1
 * eps - accuracy (1e-5)
 * - Mai99 - iteration stops to late
 * - Mar04 - rarely iteration stops to early
 *  recommend pew < NEW-1
 * out: ewc - list of converged ergs
 */
double ewc[NEW];
int convergence(zahl *ew, zahl *et, zahl eps, int itr){	/* see inc1() */
  int i1, i2, i3, i4=0, num_conv=0, in=min(max(1,pew),NEW);
  for (i3=i1=0;i1<min(itr-1,NEW);i1++){ /* check all values */
    for(i2=i3;i2<=i1;i2++)
      if(fabs(et[i2]-ew[i1])<=fabs(eps*ew[i1])){ /* found same value */
        ewc[num_conv]=ew[i1]; /* store converged values only for awk/grep */
        num_conv++;   /* count convergent values */
        i3=i2; i4=i1; /* avoid looking twice */
        break;
      } if (i1>i4+4) break; /* to much nonconverged values */
  }
  for(i1=num_conv;i1<NEW;i1++) ewc[i1]=ew[i4]; /* fill with last converged value */
  for(i1=0;i1<NEW;i1++) et[i1]=ew[i1];
  /* ToDo: save converged values in ew_conv? for better awk/grep of E0 E1 ... */
  if (!mpi_i) printf(" cnv= %d %d",num_conv,i4);
  if (i4==NEW-1 && num_conv<in && !mpi_i)
    printf ("\nWarning: (NEW) to small to get (pew) values.");
  return (num_conv>=in || i4==NEW-1);
}

void fulltridiag(unsigned int n1,zahl ta[], zahl tb[]);

/* --- --- get smallest eigenvalues (ew[]) using Lanczos --- */
/* --- Inp: iv  Out: a,b,itr,ew  Wrk: v0,v1 ret: Iterationen */
int inc1(int nev,int ne0,tvec *v0,tvec *v1){
    int i, k; long l1;
    zahl a1, b1, eps, et[NEW]; cplx t1; tvec *vtmp;
    if (vvv && !mpi_i)
      printf("\n inc1-lanczos ne0=%d (tmp/tri.txt)", ne0);
    clrvec(v0);
    if (vvv && !mpi_i) printf(" t+= %s",sclock(~0l));
    if (!mpi_i)
      printf("\n start iteration: E0..E%d i i100.t[min] num_conv last_conv",NEW-1);
    for (k=0;k<NEW;k++) ew[k]=et[k]=13.5;
#if VecType & 1
    eps=1e-11;/* high precision 1e-11 */
#else
    eps=1e-9;/* low precision */
#endif
    MPI_Barrier(MPI_COMM_WORLD); /* sync outputstreams DBG */
    l1=itime();
    if (checkpoint<4)
    {
     for(i=1;i<=MAX_ITR;i++){  /* ToDo: start with 0 (fortran=1) */
#ifdef DEBUG
       if (vvv&4 && i<4)
       { int blk,j;
          for(blk=0;blk<pt_n;blk++)
           for (j=0;j<b_len[blk+mpi_i*pt_n] && j<100;j++)
             printf("\n blk.idx= %2d %4d  i= %8lu v0[i]=%12.5g v1[i]=%12.5g",
                blk+mpi_i*pt_n, j, b_ofs[blk+mpi_i*pt_n]+j,
                v0->v[blk][j],
                v1->v[blk][j]);
         MPI_Barrier(MPI_COMM_WORLD);
       }
#endif
       /* ToDo: do the normalization within H*v operation! v1[]*=r1;v0[]*=r0 */
       /* t1=<v1*H*v1>,v0=H*v1+v0, <v1|v0>==0 !def! */
       t1=hamilton(v1,v0);      /* #FLOP see hamilton_geth_block */
       ta[i-1]=a1=Re(t1);b1=0;	/* a1=<v1*H*v1> */
       b1=add_av(v0,-a1,v1);    /* v0+=-a1*v1, b1=<v0|v0> FLOPS=4*n1 */
       tb[i-1]=b1=sqrt(b1);     /* if b1 is 0 or near 0, we should stop */
       /* or store normalization factor seperately ? */
       if (b1>0) { /* include this in hamilton like v1=H*(v0*=a)+(v1*=b) */
          mltply_av(1.0/b1,v0); /* v0=1/b1*v0 FLOPS=1*n1 */
          mltply_av(   -b1,v1); /* v1= -b1*v1 */
       }
       vtmp=v0;v0=v1;v1=vtmp; /* faster change using pointer */
       if (!mpi_i) outtri(i,a1,b1);
       if (sig1==2) { /* save before possible crash */
         savevec(v0,"tmp/v0.dat",-1,-1);
         savevec(v1,"tmp/v1.dat",-1,-1); printf(" vx saved, it=%d",i);
       }
       IFcplx(if(fabs(t1.imag())>1e-6)E1("Im<H>=% .0e",D(t1.imag()));) /* flt=1e-6 */
       /* "(i>19)"  removed, we can neglect the CPU-time for bisec() */
       if ( (i>0 && (i%5)==0)          /* convergence test all 5 iterations */
         || (ULong)i>=n1 || b1<1e-15 /* also stop on small systems */
         || n1<10 ) {                /* control convergence on small systems */
          /* compute the eigenvalues of the tridiagonal matrix T=(tb,ta,tb) */
          bisec(i,NEW,1e-13); 
          if (!mpi_i) { nl(stdout); outew(stdout);
            /* move to tri.txt for better debugging using diff? */
            printf(" i=%3u i100=%4.0f",i, (((itime()-l1)*100+i/2)/i)/60.);
          }
          if (b1<1e-15) {if(!mpi_i)printf(" inc1_b<1e-15"); break;}
          if (convergence(ew,et,eps,i)) break;
          if ((ULong)i>=n1) break;
          if (i>=MAX_ITR) {if(!mpi_i)printf(" no convergence"); break;}
       }
      }
      MPI_Barrier(MPI_COMM_WORLD); /* sync outputstreams DBG */
      if (!mpi_i) outtri(-1,0,0); /* close files */
      
      checkpoint++; if (!mpi_i) printf("\n checkpoint=%d", checkpoint);
    } else { /* reload old values */
      i=loadtri(ta,tb,MAX_ITR);
      bisec(i,NEW,1e-13); if (!mpi_i) {nl(stdout); outew(stdout);}
      if (!mpi_i) fprintf(stdout," i=%3u loaded",i);
    }
    /* ToDo: vec12 should be called by incv1, because output is needed only there
     *   for n-fold degeneracy: we need max_ea>=n, nev>=n and e0 n-fold in ew[]
     */
#if 0
    fulltridiag(i,ta,tb); /* finds exactly the same eigenvalues! */
#endif
    if (!mpi_i) { /* output converged energies for easier awk handling v2.27 */
      printf("\n conv=");
      for (k=0;k<NEW;k++) printf(" %12.8f", ewc[k]);
      printf("\n i100.t= %s i=%3u", sclock(((itime()-l1)*100+i/2)/i), i);
      if (vvv) printf(" t+= %s",sclock(~0l));
    }
    if(nev) vec12(i,nev,ne0); /* get tridiagonal eigenvectors */
    return(i);
}
    
/* --- eigenvector by the Lanczos method (after vec12())
 * --- convert eigenvectors of tridiagonal matrix back to H space 
 * same as inc1, but now we have eigenvalues (c) of triag (a,b)
 *               and do a an aditional ev[k]+=c[i][k]*v0[i]
 * --- Inp: a,b,c --- Out: v0,v1,ev=linear_access(3+itr) */
void incv1(int nev, tvec *v0, tvec *v1, tvec ev[], int itr) {
    ULong ii=1l; int i,k; long l1; zahl a1,b1,rb,rc; tvec *vtmp;
    if (nev<1) return;
    if (!mpi_i) printf("\n incv1-lanczos");
    clrvec(v0);
    for (k=0;k<nev;k++) {
      copy_av(&ev[k],tc[0][k],v1); /* ev[k]=tc[0][k]*v1 */
    }
    l1=itime();
    for(i=1;i<itr;i++,ii++){ /* v1.7.8 99% time spend for hamilton() */
       hamilton(v1,v0); /* v0=H*v1+v0, ret=<v1|H*v1> faster if v0=r1*H*v1+r2*v0 ???  */
       a1=ta[i-1]; b1=-tb[i-1]; rb=-1.0/b1; rc=tc[i][0];
       add_av(v0,-a1,v1);            /* v0+=-a1*v1 [<v0|v0>] */
       mltply_av(rb,v0);             /* v0*=-1./b1 */
       mltply_av(b1,v1);             /* v1*=    b1 */
       for (k=0;k<nev;k++)
         add_av(&ev[k],tc[i][k],v0);  /* ev[k]+=rc[k]*v0; [nevk=<evk|evk>] */
       vtmp=v0;v0=v1;v1=vtmp; /* faster change using pointer */
    }
    for(k=0;k<nev;k++) normvec(&ev[k]); /* needed? */
    if (!mpi_i) {
      printf("\n v100.t= %s i=%3d", sclock(((itime()-l1)*100+itr/2)/itr), itr);
      if (vvv) printf(" t+= %s",sclock(~0l));
    }
    return;
}
/* ------------------------------------------------------------- */

/* --- all what be needed for pthreads of diag2 --- */
typedef struct sp_diag2 {
#ifdef CONFIG_PTHREAD
  pthread_t pt;
#endif
  thbuf *hbuf;
  int *ic;
  double vv,e1,vhv,vv2,vhv2,krit;
} tp_diag2;

/*
    void-function for pthreads (one vector iteration)
    
    at start vhv is estimated
    store vhv_old=vhv, calculate vhv_last and vhv=vhv_new
    corrected vhv = (vhv_new-vhv_old)+(vhv_new-vhv_last) ???
 */
void diag2_thread( void *ptr ){
    tp_diag2 *p=(tp_diag2 *)ptr;
    tvec *v0; mcplx vnew,vold;
    cplx b,c,nv,nhv,b1,r;
    double e1,r1,nhn,vhv,vv,krit=0;
    int ic,blk; ULong n;
    blk=p->hbuf->block;
    e1 =p->e1;
    vhv=p->vhv;
    vv =p->vv;
    ic =*(p->ic);
    v0 =p->hbuf->v0;  /* |v> */
    for(n=b_ofs[blk];n<b_ofs[blk]+b_len[blk];n++)
    {    /* like MC and fixed walk path? */
       vold=v0->v[blk][n-b_ofs[blk]];      /* float, double ... */
       nv=M_f2d(vold);                     /* <nv> n-th component of |v> */
       /* dont use disk, because handling of async IO is difficult v2.26 */
       /* nhn: <nHv> n-th component of H|v> */
       /* nhv: <n|Hn> diagonal-H of |n> (including <nHn>???) */
       nhv=  hamilton_nhv(n-b_ofs[blk],p->hbuf,0);
       nhn=  p->hbuf->nhn;
       // if (nn==6) printf(" n nhn %lu %f\n", n, nhn );
       /* nhn to small for sym until v2.25 */
       /* ToDo: can iteration be splitted into sub-H-blocks? H1,H2,H3 etc?
        *       if not, we have to write paramindex per nonzero element */ 
       if(ic==0){ vhv+=D(Conj(nv)*nhv); continue; } /* only calculate vhv_last */
       /* ----------------------- matrix element of S^(-1)H*(1-nv*nv)  */
       b=vv*nhv-vhv*nv; /* a=vv*nhn-nv*nhv; */ if(fabs(b)<1e-30)continue;
       c=   nhv-nhn*nv; /* d=   vhv-nhv*nv; */
       /* <b|c>==vv*<nhv|nhv>+vhv*nhn*<nv|nv>-vv*nhn*<nhv|nv>-vhv*<nv|nhv> ??? */
       r1=vhv-vv*nhn;   /* d-a */
       if(fabs(b*c)<1e-12){continue;}  /* no changes in v0 */
       r =r1-sqrt(r1*r1+4*b*c);	/* complex??? ew r1^2-4bc=(a+d)^2-ad-bc=(p/2)^2-q */
       b1=2*b/r;                /* correction b1 is small??? */
       vnew=vold+M_d2f(b1);     /* rounding error ??? vold==nv */
       b1=M_f2d(vnew-vold);     /* set b1 according to rounding error */
       v0->v[blk][n-b_ofs[blk]]=vnew;
       krit+=Norm(b1);	/* Norm(old.v-new.v), better use max(fabs(b1))??? */
       /* vv =<v+b|v+b>  =<v|v>  +2<b|nv>  +<b|b> */
       /* vhv=<v+b|H|v+b>=<v|H|v>+2<bn|Hv>+<b|H|b> */
       /* not ok! complex ????????? */
       vv  +=D(2*Conj(b1)*nv )+Norm(b1);     /* correct besides matherrors */ 
       vhv +=D(2*Conj(b1)*nhv)+Norm(b1)*nhn;	/* new EW allways lower! */
       /* bad startvec can lead to overflows or big_summed_error */
       if( vv<0.5 || vv>1.6 || fabs(vhv/vv-e1)>4 ){
         printf(" warning: dE=%e vv=%f b1=%f\n",fabs(vhv/vv-e1),vv,D(b1));
         /* *(p->ic)=-1; */ 
         break;
       }
    } /* for(n) */
    p->krit=krit;
    p->vhv2=vhv;
    p->vv2 =vv;
}

/* --- --- get smallest eigenvalues using special-Lanczos ----------
 *   see 2x2.tex for explanations 
 * --- Inp: iv,v0=startvector  Out: itr,ew,v0=eigenvector  ---------- */
int lanz2(tvec *v0){
    double r,e1,vhv,vv,eps=1e-9,krit;
    int itr,ic,blk,j; ULong /* i, */ ti;
    tp_diag2 pv[B_NUM],*p;
    
    for (ic=0;ic<pt_n;ic++) pv[ic].hbuf=alloc_hbuf(); /* buffer space */
#if (VecType & 3) == 0
    eps=1e-8;  		/* float JS-Apr99 */
#endif        
#if (VecType & 8)
    eps=1e-4;  		/* float16 JS-Feb02 find more elegant way! */
#endif        
    printf(" lanz2: eps=%5.0e\n",eps);
    vhv=0.0; e1=13.5; vv=1.0; ti=itime();
    for(itr=1,ic=0;itr<4*MAX_ITR;itr++,ic=(ic+1)&31){
       /* ToDo: reset ic on external signal sig1==1 ??? */
       krit=0;
       if (ic==0) { vhv=0; krit=999; }
       for(blk=0;blk<pt_n;blk++){ p=&(pv[blk]);
         p->hbuf->v0   =v0;
         p->hbuf->block=blk;
         p->vv         =vv;
         p->vhv        =vhv;
         p->e1         =e1;
         p->ic         =&ic;
         parallel_start(blk, diag2_thread, p);
       }       
       for(blk=0;blk<pt_n;blk++){ p=&(pv[blk]);
         parallel_wait(blk);
         vv  += p->vv2  - p->vv;
         vhv += p->vhv2 - p->vhv;
         krit+= p->krit;
       }
       r=1/sqrt(vv);  mltply_av(r,v0); /* v0*=r */
       krit*=fabs(r); krit=sqrt(krit/(double)n1);
       e1=vhv/=vv; vv=1.0;		/* normieren */
       /* if(!(itr%5)) */
       {
         printf("%3d %20.15f %5.0e %4s %2d %6.0e\n",
                itr,D(e1),(double)krit,sclock(~0l),ic,1-r);
         if (krit<eps) break;
       }
       /* here the user can stop iterations (ex: slow convergence) and 
        * continue with saving vector and do the next steps,
        * usefull if computation must be stopped and proceeded later
        * since v2.26
        */
       if (sig1 == 2) { /* signal 2*USR2+1*USR1, stop and save vector */
         sig1=0; /* reset signal */
         printf("signal 2 received, stop iteration\n");
         break;
       }
       if(itime()-ti>60*60*12l){ ti=itime(); /* all 12h -> save v0 */
         /* savevec(v0,"d1.vec",-1,-1); # better msync? */ }
    }
    for (j=0;j<NEW;j++) ew[j]=D(e1);
    for (ic=0;ic<pt_n;ic++) free_hbuf(pv[ic].hbuf);
    if (vvv) printf(" t+= %s",sclock(~0l));
    return(itr);
}
    
/* --------------------------  jacobi ----------------------------------- */
/*
 * jacobi does _not_ work for complex matrizes (in the present form)!
 * but for complex (hermitian) matrizes of type: A+iB
 *   use [[ A,-B], [ B, A]] as virtual matrix,
 *   in that case all eigenvalues got twice.
 *   and computing time is 9.2 times of the real time version
 *   but what about eigenvectors?
 * Driver for routine jacobi for symmetric real matrices
 * - what about parallelization? (BFG-package by I.J.Bush?)
 * sparc-Ultra1-Boson: time/s = 1.5e-6 dim ^ 3.2
 * A: n1*n1 (complex: 2*n1*n1) (ToDo)
 * D: n1    (complex: 2*n1)
 * V: n1*n1 (complex: (2*n1)^2)
 * --------------------------  jacobi ----------------------------------- */
#define ROTAT2(a,i,j,k,l) g=a[i*n+j];h=a[k*n+l];a[i*n+j]=g-s*(h+g*tau);\
	a[k*n+l]=h+s*(g-h*tau);
void jacobi2(double a[], int n, double d[], double v[], int *nrot){
    int j,iq,ip,i;
    double theta,tau,t,s,h,g,c,*b,*z,tresh,sm;
    if(!(b=(double *)malloc((size_t) (n*S_double)))) F0("malloc");
    if(!(z=(double *)malloc((size_t) (n*S_double)))) F0("malloc");
    for(ip=0;ip<n;ip++){
     for(iq=0;iq<n;iq++)v[ip*n+iq]=0.0;v[ip*n+ip]=1.0;}
    for(ip=0;ip<n;ip++){b[ip]=d[ip]=a[ip*n+ip];z[ip]=0.0;}
    *nrot=0;
    for (i=1;i<=50;i++){
      sm=0.0;
      for(ip=0;ip<n-1;ip++)for(iq=ip+1;iq<n;iq++)sm+=fabs(a[ip*n+iq]);
      if(sm==0.0){ break; } /* finished */
      if (i < 4)tresh=0.2*sm/(n*n);
	else    tresh=0.0;
      for (ip=0;ip<n-1;ip++) {
	for (iq=ip+1;iq<n;iq++) {
	 g=100.0*fabs(a[ip*n+iq]);
	 if (i>4 && (fabs(d[ip])+g)==fabs(d[ip])
		 && (fabs(d[iq])+g)==fabs(d[iq])) a[ip*n+iq]=Z(0.0);
	 else if (fabs(a[ip*n+iq]) > tresh) {
	  h=d[iq]-d[ip];
	  if ((fabs(h)+g)==fabs(h)) t=(a[ip*n+iq])/h;
	  else { theta=0.5*h/(a[ip*n+iq]);
		 t=1.0/(fabs(theta)+sqrt(1.0+theta*theta));
		 if (theta < 0.0) t = -t;}
	  c=1.0/sqrt(1+t*t);s=t*c;tau=s/(1.0+c);h=t*a[ip*n+iq];
	  z[ip] -= h;z[iq] += h;d[ip] -= h;d[iq] += h;a[ip*n+iq]=Z(0.0);
	  for (j=0;j<=ip-1;j++)    { ROTAT2(a,j,ip,j,iq)}
	  for (j=ip+1;j<=iq-1;j++) { ROTAT2(a,ip,j,j,iq)}
	  for (j=iq+1;j<n;j++)     { ROTAT2(a,ip,j,iq,j)}
	  for (j=0;j<n;j++)        { ROTAT2(v,j,ip,j,iq)}
	  ++(*nrot);
	 }
	}
      }
      for (ip=0;ip<n;ip++){ b[ip]+=z[ip]; d[ip]=b[ip]; z[ip]=0.0;}
    } 
    if(i>50) E0("Too many iterations in routine jacobi");
    free(z); free(b);
    return;
}
#undef ROTAT2
/* ----------------------------------- end jacobi ---------------------- */

void shellsort(double a[],int ia[],ULong n);
/* call jacobi to diagonalize tridiag
 * --------------------------  Matrizen ----------------------------------- */
void fulltridiag(unsigned int n1,zahl ta[], zahl tb[]){
    int nrot,*ix; ULong i,j;
    double *xm,*xv,*xd,xi;     /* Matrizen 1 & 2 */
    if(vvv)printf("\nfulltridiag"); xi=0;
    /* standart jacobi and LAPACK expects simple vectors, convert block form */
    /* add 1 for the case n1=1, where Lapack fails */
    xm=(double *)mymalloc(n1*n1*S_double,NULL); /* Matrix_H-in */
    xv=(double *)mymalloc(n1*n1*S_double,NULL); /* eigenvects-out */
    xd=(double *)mymalloc(   n1*S_double,NULL); /* eigenvalues-out */
    ix=   (int *)mymalloc(   n1*S_int,NULL);    /* sort-idx-out */
    for(i=0;i<n1;i++) ix[i]=(int)i; /* initialize sort-index (LAPACK) */
    if(xm!=NULL && xv!=NULL && xd!=NULL && ix!=NULL){
      if (vvv) printf(" H");
      /* generate a simple matrix, store all elements to xm[] */
      for(i=0;i<n1;i++) for(j=0;j<n1;j++){ xm[i*n1+j]=0.0; }
      for(i=0;i<n1;i++) xm[i*n1+i]=ta[i]; 
      for(i=0;i<n1-1;i++) { xm[i*n1+i+1]=tb[i]; xm[(i+1)*n1+i]=tb[i]; }
      printf(" JACOBI");
      jacobi2(&xm[0],(int)(n1),xd,xv,&nrot);
      printf(" rotations: %3d",nrot);
      shellsort(xd,ix,n1);		/* sort xd[ix[i]] */
      printf("\n eigenvals=");
      for (j=0;j<n1;j++){ printf(" %+12.8f",xd[ix[j]]); } printf("\n");
    }
    myfree((void *)ix,NULL);
    myfree((void *)xm,NULL);
    myfree((void *)xv,NULL);
    myfree((void *)xd,NULL);
    return;
}
