/*
    Jun08,2000: sleep in mymalloc added, if temporarely not enough memory
    can give problems if parallelized!
    
 ToDo: 
       - using fd=mkstemp((char *)s1="tmp/mmXXXXXX") remove(s1);
       - replace mymalloc(...,NULL) calls
       - replace mymalloc, myfree by remalloc(0,0,...)

  v1.2 - fallback from malloc to mmap added (tmpfile if not enough mem) 

 */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h> /* sleep() */
#include "defs.h" /* ULong */
#include "mem.h"

/* some external vars, not needed to include big h-files */
extern ULong usemem,maxmem;
extern int vvv;   /* verbose mode */
extern int mpi_i; /* mpi node */

minfo_t test0 = { NULL, 0, 0, "tmp/t_XXXXXX", NULL };

/* return 0 on success (old was: manipulating void **pp) */
void *myfree(void *pp, minfo_t *mi){ 
    if (!pp) return pp;
    if (mi) {
      usemem-=min(mi->l,usemem);
      if (vvv&2 && !mpi_i) printf("\n  free(%.0e)=%10p -= %3.1e => %s",
               (double)mi->l,pp,(double)usemem,((mi)?mi->name:""));
      if (pp!=mi->p) E0("mi.p != pp");
      if (!mi->l)    E0("mi.l == 0");
    }
    if (mi && mi->ff) {
      myunmap( pp,mi->ff,(long)mi->l,0);
      fclose(mi->ff); mi->ff=NULL;
    } else {
      free(pp);
    }
    if (mi) { mi->p=NULL; mi->l=0; }
    return NULL;
}

void *myremalloc(void *p0,unsigned long ll, minfo_t *mi){
    int itry; void *pp=NULL;

    if (ll==0l) { return myfree(p0,mi); }

    if (mi && !mi->p) mi->l=0;
    
    if ( (!mi) || (!mi->ff) )                  /* not mmapped */
    if ( usemem+((mi)?ll-mi->l:0) <= maxmem
      || maxmem==0l || (mi && (mi->mode&4)))
    for(itry=1;itry<5 && pp==NULL;itry++){    /* try for 120m=2h */
      pp = (void *)realloc(p0,(size_t)ll);
      /* if realloc failed? ret==0 => data untouched (not freed) */
      if (pp!=NULL || maxmem==0l) break;
      printf("\n alloc( %10p,%.0e) => %s failed, sleep & repeat",
             p0,(double)ll,((mi)?mi->name:""));
      sleep(itry*120); /* 120sec */
    }
    if (mi) {
#ifdef HAVE_MMAP
      if (!pp) { /* try mmap */
        if (!mi->ff) {
          mi->ff=fopen(mi->name,"r+b"); if(!mi->ff)
          mi->ff=fopen(mi->name,"w+b"); /* create */
          if (p0 && mi->l) 
           if (fwrite(p0,1,mi->l,mi->ff)!=mi->l) F1("write %s",mi->name);
          if (p0) free(p0);
          if (mi->ff) { pp=mymap(mi->name,mi->ff,(long)ll,0); }
          if (pp) { mi->l=ll; }
        } else { /* remap */
          itry=vvv; vvv=0; /* be quiet */
          myunmap( p0      ,mi->ff,(long)mi->l,0); mi->l=0;
          pp=mymap(mi->name,mi->ff,(long)ll,0);
          vvv=itry;
        }
      }
#endif
      if (pp && !mi->ff) usemem += ll - mi->l;
      if((vvv&2) && p0==NULL && !mpi_i) {
        /* %11.6g means print 6 digits number or (6+5) e-number */
        if (usemem>9999999)
          printf("\n alloc(%7.1e)=%10p += %7.1e => %s node= %2d",
                 (double)ll,pp,(double)usemem,((mi)?mi->name:""), mpi_i);
        else
          printf("\n alloc(%7.0f)=%10p += %7.0f => %s node= %2d",
                 (double)ll,pp,(double)usemem,((mi)?mi->name:""), mpi_i);
      }
      if (pp) { mi->p=pp; mi->l=ll; }
      if (!(mi->mode&1)) return pp;
    }
    if (!pp) { F2("realloc failed(%p,%ld), aborting",p0,(long)ll); }
    return pp;
}
    
void *mymalloc( unsigned long ll, minfo_t *mi){
    return myremalloc(NULL, ll, mi);
}

/* -------------------------------------------------------------------
    using files as memory to handle arrays bigger than mem+swap
    see libc.info (Memory-mapped I/O) for more information
    see utils/mmap1.c too
    how to use LFS (64bit)?
   ------------------------------------------------------------------- */
#define DEBUG 1

#ifdef HAVE_MMAP      /* else use malloc */
#include <unistd.h>   /* lseek (off_t = 64bit on LFS) */
#include <sys/mman.h> /* mmap */

static size_t page_size=0;
#endif

/*  map a memory to a file */
void *mymap(char *nam, FILE *f1,long len,long ofs){
  int fd=-1,dummy=0; long i,pos; void *buf;
  if(len==0) return NULL;
#ifdef HAVE_MMAP
  fd=fileno(f1);
  /* systems have very different demands for offset in mmap function */
  /* here I try to find best for all systems */
#ifdef CFG_PAGESIZE /* this is not defined on every system (iMAC?) */
  page_size= (size_t) sysconf(_SC_PAGESIZE);
#else   /* if trouble: ./configure --mmap */
  page_size= 8192;  /* should work mostly, best (2^n) : n>=14 */
#endif
#endif
  fseek(f1,(size_t)0,SEEK_END); /* fseek after file end failes */
  pos=ftell(f1);
  for(i=0;i<ofs+len-pos;i++) fwrite(&dummy,1,1,f1); /* fill up */
  fflush(f1); /* expand file */
  /* the file must have already the size of the map and opened for r/w */
#ifdef HAVE_MMAP
  /* ofs must be a multiple of getpagesize() !!! */
  /* but most system dont care about that, on trouble do: configure --nomap */
  /* ToDo: use CFG_PAGESIZE defined by configure to fix that problem */
  if( page_size<=(size_t)len )
    buf=mmap(NULL,(size_t)len,PROT_READ | PROT_WRITE,MAP_SHARED, fd, (off_t)ofs);
  else 
#endif
  {
    buf=malloc((size_t)len); fd=-1;
    if(buf==NULL) E1("malloc(len=%.0e)",(double)len);
    fseek(f1,(long)ofs,SEEK_SET);pos=(long)fread(buf,1,len,f1);
    if( pos!=len ) E0("my_map: fread failed");
    fseek(f1,(long)0,SEEK_END); /* fseek after file end failes */
  }
  if(vvv&2 && !mpi_i){ /* fd=-1 if not mapped but loaded to memory */
    if (ofs<99999l && len<99999l)
      printf("\n mmap %s ofs=%5.0f len=%5.0f to %p fd=%2d node=%2d",nam,1.*ofs,1.*len,buf,fd,mpi_i);
    else
      printf("\n mmap %s ofs=%5.0e len=%5.0e to %p fd=%2d node=%2d",nam,1.*ofs,1.*len,buf,fd,mpi_i);
  }
  /* on error fall back to malloc+load (save+free) */
  if(buf==(void *)-1) { E3("mmap(fd=%d,len=%ld,ofs=%ld)",fd,(long)len,(long)ofs); perror(" "); }
  return buf;
}

int myunmap(void *buf, FILE *f1, long len, long ofs){
  int rc=0; long pos;
  if (buf==NULL || buf==(void *)-1 || len==0) return 0;
  if (vvv&2 && !mpi_i)
    printf("\n unmap %p len=%.0e ofs=%.0e",buf,(double)len,(double)ofs);
  if (!f1)  E0("no open FILE");
#ifdef HAVE_MMAP
  if ( page_size<=(size_t)len ){
    rc=msync((caddr_t)buf,(size_t)len,MS_SYNC);
      if (rc) fprintf(stderr,"ERROR msync=%d\n",rc);
    rc=munmap((caddr_t)buf,(size_t)len);
      if (rc) fprintf(stderr,"ERROR munmap=%d\n",rc);
  } else
#endif
  {  /* save! */
    if (vvv&2 && !mpi_i) printf(" write_to_disk");
    fseek(f1,(long)ofs,SEEK_SET); pos=(long)fwrite(buf,1,len,f1);
    if ( pos!=len ) E0("my_unmap: fwrite failed");
    if (buf) free(buf);
  }
  return rc;
}
