/*
  linux: mpicc -O2 -g -Wall ../utils/mpi_speed.c
  P4-1.3GHz-linux 15.3us      476MB/s    (2)
  ES45-Tru64-5.1B  2.6us 5%   820MB/s    (2)
                   2.79us    1130MB/s 2% (4)
      overload-2x 26.0us 4%   7.2MB/s 0.2%(8) 
  GS1280-32:       3.1us 4%   715MB/s 4% (2)
                   3.9us 3%  1960MB/s(512kB)
                             1560MB/s(1MB) 3% (8)
                   6.4us 4%  3930MB/s(32kB)
                             4040MB/s(64kB)
                             4040MB/s(128kB)
                             4160MB/s(256kB)
                             3500MB/s(512kB)
                             2500MB/s(1MB) 3% (32) 
  Altix330:        2.3us 3% 14900MB/s(512kB) 9160MB/s(1MB) 3% (2)
                   3.2us 4% 12500MB/s(512kB) 5100MB/s(1MB) 4% (4) 
  2DualAMD265:      23us 4%   660MB/s 4% (last!)   baldur
  4Xeon3GHz_tina   3.3us 4%   307MB/s 4% (4) Dec07 node004
    2*1G via .1G   160us 5%    15MB/s 3% (2) Dec07 node4+node012
                   191us 4%    28MB/s 3% (4) Dec07 node4+node012
                   206us 4%    63MB/s 3% (8) Dec07 node4+node012

*/
#include <stdlib.h>
#include <stdio.h>
#include <mpi.h> 
#include <time.h>

#define BUFSIZE (512*1024)
int mpi_n, mpi_i;
char buf1[BUFSIZE], buf2[BUFSIZE];
static int header=0;

int testrun(unsigned long len){
  unsigned long i1, i2, i3;
  int node1, node2; time_t t1, t2;
  MPI_Status status;
  if (!header) { header=1;
    if (!mpi_i)
    printf("#    msgsize[B] loops[2^x] time[s]"
           " latency[us] SumBW[MB/s] error\n"); /* BW=bandwidth*/
  }
  i3=8;
  for (i1=1l<<i3;i1;i1<<=1,i3++){
    time(&t1);
    for (i2=0;i2<i1;i2++) {
      node1=(mpi_i-(i2%(mpi_n-1)+1)+mpi_n)%mpi_n;
      node2=(mpi_i+(i2%(mpi_n-1)+1)+mpi_n)%mpi_n;
      MPI_Sendrecv(&buf1,len,MPI_BYTE,node1,0,
                   &buf2,len,MPI_BYTE,node2,0,MPI_COMM_WORLD,&status);
    }
    time(&t2); t2-=t1;
    if (t2>1 && !mpi_i)
    printf("SumNodes %8lu %6lu %8d %9.2f  %8.2f  %5.1f%%\n",
           len, i3, (int)t2, t2*1.0e6/i1, i1*1.0e-6*len*mpi_n/t2, 100./t2);
    if (t2>20) break;
  }
  if (!mpi_i)
  printf("SumNodes %8lu %6lu %8d %9.2f  %8.2f  %5.1f%% <= final\n",
           len, i3, (int)t2, t2*1.0e6/i1, i1*1.0e-6*len*mpi_n/t2, 100./t2);
  return 0;
}


int main(int argc, char *argv[]) {
  unsigned long i1;
  setvbuf(stdout,NULL,_IONBF,0);
  if (MPI_Init(&argc, &argv) != MPI_SUCCESS) exit(1);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_n);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_i);
  { int version, subversion;
  MPI_Get_version(&version,&subversion);
  if (!mpi_i)
   printf(" using MPI %d.%d with %d nodes\n", version, subversion, mpi_n);
  }
  for (i1=0;i1<BUFSIZE;i1++) buf1[i1]=buf2[i1]=(char)i1;
  MPI_Barrier(MPI_COMM_WORLD);
 
  if (!mpi_i) printf("Measure speed of MPI_sendrecv transactions:\n");
  testrun(1);
  testrun(2);
  testrun(BUFSIZE/2);
  testrun(BUFSIZE);
  
  MPI_Finalize();
  return 0;
}
